/*.............................................................

Project: Tiberium
Version: 1.4
Author: NLD

Date: 2021/Apr/1

ESP8266 based wireless current monitor and smart plug.

This code might not compile for other ESP8266 models than the LOLIN(WEMOS) D1 R2 & Mini !
Also tested on: NodeMCU1.0 (ESP-12E Module).


Build settings:
 -Use upload/serial: 115200 
 -Use CPU/ram: 160Mhz/4M flash (don't use 80Mhz cpu LoL)
 -Use FLASH size: 4M (2M OTA SPIFFS)
 -Use erase sketch + wifi

Upload via OTA:
 - Choose Sketch/Export compiled binary to generate the .bin file
 - Use the following command to upload it to the ESP8266:
 
 python <LOCATION OF YOUR ESP LIBRARY>/espota.py -p 8266 -i <PLUGIP> -a "<OTA_PASSWORD>" -f <OUTPUT BIN FILE>

 

DO NOT Power ACS from other pins than the ESP8266 5V input rail, the voltage reference changes!

Changes:
v1.4:
 - Rapidly polling the ADC for the power/current calculations caused the malfunction in the ESP's Wifi stack which lead to lag spikes/packet loss
 and other issues therefore the stats polling is now limited to once in 10 seconds, putting this on throttle also made button pushes instant
 - OTA code is now optional, most of the time it does not work anyway, fails to connect, fails to upload other issues
 - The average response times after these fixes:
  HTTPs m2m 0.6 seconds
  HTTP  m2m 0.1 second
v1.3:
 - Bugfix in the webserver code, in case of m2m communication the condition fall through and it provided part of the regular page as well
v1.2:
 - Net_init() added
 - Wifi reconnect delay increased
 - Due to some memory corruption on the longrun an option added to run regular HTTP server instead of HTTPS
v1.1:
 - Version added
 - ESP8266 Community library upgraded from 2.6.3->2.7.1
v1.0:
 - OTA updater fixed
 - Periodic wifi connection check added
v0.9:
 - Automatic restart added if it cannot connect to the wifi 3 times
v0.8:
 - Restart command have been added as a workaround to reset the device periodically for stable operation
v0.7:
 - LED color purple added to differentiate between the wifi connect/ntp errors. Ledcolor 5 now means NTP error while 6 means off.
v0.6:
 - Code updated to ESP8266 Arduino library 2.6.2 (will not compile with earlier versions anymore due to syntax changes)
 - OTA updates added
v0.5:
 - Webserver part rewritten for chunked transfer
 - PowerAVG sampling increased from 100->200
 - Autoredirect back to root page after saving time entries
 - Time entries corrected, max hh 23, max mm 59
 - All temp variables, testcodes were cleaned up
 - NTP library placed inside the project
 - Staic IP code added
 - Timer can be turned ON/OFF now dynamically any time instead of being a build option
 - Option for disabling the state save to flash was removed
 - Timer clean code added
 - Sample cert regenerated with 1024 bits
 - Initial NTP sync check added, if it can't connect to the NTP server the plug will try forever since it's a vital function
 - Ledcolor fix afer failed wificonn, ntpnego (yellow->actual color)
 - Instead of ON/OFF, the reason of the last ON/OFF event is also displayed on the webpage
v0.4:
 - ON/OFF states changed
 - Timer programming api through serial/web implemented
v0.3:
 - Time switch functionality implemented
 - Amps correction changed from -0.04 to -0.12 A
 - Without load the ACS is generating flickering output therefore all almp values under 0.005A are ignored
v0.2:
 - Uptime added
 - LED/switch pins changed
v0.1:
 - Based on EoL v0.1

Todo:
 - Add automatic adjustment for winter/summer time. Right now the offset needs to be replaced every time and the code reuploaded. 
..............................................................*/
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <WiFiUdp.h> 
#include <ArduinoOTA.h>
#include "src/NTPClient/NTPClient.h"

//#define SECURE // HTTP(s) will slow down the connection. You decide if you need it on a LAN...
//#define OTA   // OTA is kinda broken anyways, only turn on if you need it
#define VERSION "1.4"   // Set the software version
#define NODENAME "R6"   // Set the node name 
//#define DEBUG           // Debug mode (for the fastest production use, disable this)
#define STATIC_IP       // If disabled the plug will get it's ip from the DHCP


/* Wifi Details */	
#define WIFI_TIMEOUT 30000              // checks WiFi every ...ms. Reset after this time, if WiFi cannot reconnect.
char ssid[] = "<youraccesspoint>";      // your network SSID (name)
char pass[] = "<yourpassword>";         // your network password 

/* Static IP address configuration */
IPAddress staticIP(10, 0, 0, 2);     //ESP8266 static ip
IPAddress gateway(10, 0, 0, 1);      //IP Address of your WiFi Router (Gateway)
IPAddress subnet(255, 255, 255, 0);  //Subnet mask
IPAddress dns(10, 0, 0, 1);          //DNS


/* WEB */
#ifdef SECURE
#include <ESP8266WebServerSecure.h>
#else
#include <ESP8266WebServer.h>
#endif

/* NTP 
Just a note on the offset: 
7200 => GMT + 2Hr is SUMMER TIME
3600 => GMT + 1Hr is WINTER TIME
*/
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, "10.0.0.1", 3600, 86400000); // Offset GMT +2, Update interval 24h 86400000

/* TIMER */
unsigned int timer_enabled=0;
unsigned int t1_set, t1_hh, t1_mm;
unsigned int t2_set, t2_hh, t2_mm;
unsigned int t3_set, t3_hh, t3_mm;
unsigned int t4_set, t4_hh, t4_mm;
unsigned int t5_set, t5_hh, t5_mm;
unsigned int t6_set, t6_hh, t6_mm;
unsigned int t7_set, t7_hh, t7_mm;
unsigned int t8_set, t8_hh, t8_mm;
unsigned int t9_set, t9_hh, t9_mm;
unsigned int t10_set, t10_hh, t10_mm;

/* Push button */
int buttonState;             // the current reading from the input pin
int lastButtonState = LOW;   // the previous reading from the input pin
unsigned long lastDebounceTime = 0;  // the last time the output pin was toggled
unsigned long debounceDelay = 50;    // the debounce time; increase if the output flickers

/* Webserver */

/*
 * * * HOW TO GENERATE YOUR OWN CERTIFICATE/KEY PAIR * * *
    A sample script, "make-self-signed-cert.sh" is provided in the
    ESP8266WiFi/examples/WiFiHTTPSServer directory.  This script can be
    modified (replace "your-name-here" with your Organization name).  Note
    that this will be a *self-signed certificate* and will *NOT* be accepted
    by default by most modern browsers.  They'll display something like,
    "This certificate is from an untrusted source," or "Your connection is
    not secure," or "Your connection is not private," and the user will
    have to manully allow the browser to continue by using the
    "Advanced/Add Exception" (FireFox) or "Advanced/Proceed" (Chrome) link.
*/

// Example cert
static const uint8_t x509[] PROGMEM = {
  0x30, 0x82, 0x01, 0xdd, 0x30, 0x82, 0x01, 0x46, 0x02, 0x09, 0x00, 0x87,
  0x9f, 0x8b, 0x1c, 0x60, 0xa9, 0x9d, 0x9b, 0x30, 0x0d, 0x06, 0x09, 0x2a,
  0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x30, 0x33,
  0x31, 0x15, 0x30, 0x13, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x0c, 0x54,
  0x69, 0x62, 0x65, 0x72, 0x69, 0x61, 0x6e, 0x20, 0x53, 0x75, 0x6e, 0x31,
  0x1a, 0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x11, 0x69, 0x6e,
  0x73, 0x74, 0x72, 0x75, 0x63, 0x74, 0x61, 0x62, 0x6c, 0x65, 0x73, 0x2e,
  0x63, 0x6f, 0x6d, 0x30, 0x1e, 0x17, 0x0d, 0x31, 0x39, 0x31, 0x30, 0x31,
  0x38, 0x30, 0x31, 0x32, 0x33, 0x32, 0x34, 0x5a, 0x17, 0x0d, 0x33, 0x33,
  0x30, 0x36, 0x32, 0x36, 0x30, 0x31, 0x32, 0x33, 0x32, 0x34, 0x5a, 0x30,
  0x33, 0x31, 0x15, 0x30, 0x13, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x0c,
  0x54, 0x69, 0x62, 0x65, 0x72, 0x69, 0x61, 0x6e, 0x20, 0x53, 0x75, 0x6e,
  0x31, 0x1a, 0x30, 0x18, 0x06, 0x03, 0x55, 0x04, 0x03, 0x0c, 0x11, 0x69,
  0x6e, 0x73, 0x74, 0x72, 0x75, 0x63, 0x74, 0x61, 0x62, 0x6c, 0x65, 0x73,
  0x2e, 0x63, 0x6f, 0x6d, 0x30, 0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a,
  0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x81,
  0x8d, 0x00, 0x30, 0x81, 0x89, 0x02, 0x81, 0x81, 0x00, 0xdf, 0x43, 0x9a,
  0xae, 0x2f, 0x6b, 0xb7, 0x72, 0x3e, 0x74, 0xb7, 0x35, 0x14, 0xeb, 0xe0,
  0x6e, 0x43, 0xdd, 0xb6, 0xb6, 0x58, 0xb5, 0x9a, 0x0b, 0x53, 0xee, 0x4c,
  0x5d, 0x8a, 0xaf, 0xd4, 0xac, 0x84, 0x0f, 0x61, 0x1a, 0x68, 0xf7, 0xa3,
  0xe0, 0x8a, 0x90, 0x2e, 0x0c, 0xa5, 0x14, 0xb0, 0x78, 0x97, 0x68, 0x99,
  0x5c, 0xa5, 0x5b, 0x8b, 0x35, 0x11, 0xef, 0x31, 0x2d, 0x0b, 0xd9, 0xe5,
  0xbf, 0x12, 0xc4, 0x28, 0xef, 0xf4, 0x59, 0xd7, 0xb0, 0xec, 0x95, 0x58,
  0x1b, 0x3c, 0xfe, 0x1c, 0x10, 0xcb, 0x24, 0xab, 0x92, 0x02, 0x18, 0x53,
  0x87, 0xd2, 0xb4, 0x89, 0xf8, 0x6b, 0xab, 0x40, 0x82, 0x48, 0x9d, 0xc1,
  0x8f, 0x8b, 0x1e, 0x7f, 0x7a, 0xbb, 0x28, 0xdd, 0x63, 0xb0, 0x9d, 0x59,
  0xe1, 0xb0, 0xfe, 0xd6, 0x3d, 0x92, 0x92, 0x43, 0xba, 0x80, 0xb6, 0x09,
  0xa5, 0x42, 0xb9, 0x6c, 0x91, 0x02, 0x03, 0x01, 0x00, 0x01, 0x30, 0x0d,
  0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05,
  0x00, 0x03, 0x81, 0x81, 0x00, 0x5f, 0x9a, 0xad, 0xd3, 0xa7, 0x63, 0x83,
  0x36, 0x63, 0x4a, 0x91, 0xc8, 0x9c, 0xb0, 0xac, 0x64, 0xd9, 0x70, 0x57,
  0x9c, 0xfb, 0x54, 0x99, 0x6b, 0xc8, 0xf1, 0xc3, 0x7f, 0x14, 0x7f, 0x0a,
  0x0a, 0x82, 0xc5, 0x9a, 0x3a, 0xcc, 0xd5, 0x45, 0xd3, 0x0b, 0x54, 0x43,
  0x4a, 0xfc, 0x65, 0xe6, 0xa3, 0x6a, 0x42, 0x98, 0xbf, 0xd9, 0x30, 0x4b,
  0x3c, 0x23, 0xc0, 0x10, 0x3c, 0x83, 0x16, 0x5c, 0xa2, 0x51, 0x4c, 0x95,
  0xfa, 0x8a, 0x48, 0x68, 0x8e, 0xd5, 0xb6, 0x27, 0xee, 0x23, 0x5d, 0x55,
  0x43, 0xfd, 0xd4, 0x6b, 0xc0, 0x6a, 0x84, 0xdd, 0x15, 0x8f, 0x3c, 0x21,
  0x62, 0x65, 0x9f, 0x59, 0x10, 0xe1, 0x12, 0x74, 0x09, 0x33, 0xbb, 0xc8,
  0x87, 0x24, 0x6a, 0x41, 0x4a, 0x6c, 0xf7, 0xfc, 0xaf, 0xbc, 0xfd, 0xa8,
  0x3e, 0x2c, 0x33, 0xde, 0x35, 0xb8, 0xb5, 0x16, 0x55, 0xd6, 0xe7, 0xf4,
  0x38
};


// And so is the key.  These could also be in DRAM
static const uint8_t rsakey[] PROGMEM = {
  0x30, 0x82, 0x02, 0x5d, 0x02, 0x01, 0x00, 0x02, 0x81, 0x81, 0x00, 0xdf,
  0x43, 0x9a, 0xae, 0x2f, 0x6b, 0xb7, 0x72, 0x3e, 0x74, 0xb7, 0x35, 0x14,
  0xeb, 0xe0, 0x6e, 0x43, 0xdd, 0xb6, 0xb6, 0x58, 0xb5, 0x9a, 0x0b, 0x53,
  0xee, 0x4c, 0x5d, 0x8a, 0xaf, 0xd4, 0xac, 0x84, 0x0f, 0x61, 0x1a, 0x68,
  0xf7, 0xa3, 0xe0, 0x8a, 0x90, 0x2e, 0x0c, 0xa5, 0x14, 0xb0, 0x78, 0x97,
  0x68, 0x99, 0x5c, 0xa5, 0x5b, 0x8b, 0x35, 0x11, 0xef, 0x31, 0x2d, 0x0b,
  0xd9, 0xe5, 0xbf, 0x12, 0xc4, 0x28, 0xef, 0xf4, 0x59, 0xd7, 0xb0, 0xec,
  0x95, 0x58, 0x1b, 0x3c, 0xfe, 0x1c, 0x10, 0xcb, 0x24, 0xab, 0x92, 0x02,
  0x18, 0x53, 0x87, 0xd2, 0xb4, 0x89, 0xf8, 0x6b, 0xab, 0x40, 0x82, 0x48,
  0x9d, 0xc1, 0x8f, 0x8b, 0x1e, 0x7f, 0x7a, 0xbb, 0x28, 0xdd, 0x63, 0xb0,
  0x9d, 0x59, 0xe1, 0xb0, 0xfe, 0xd6, 0x3d, 0x92, 0x92, 0x43, 0xba, 0x80,
  0xb6, 0x09, 0xa5, 0x42, 0xb9, 0x6c, 0x91, 0x02, 0x03, 0x01, 0x00, 0x01,
  0x02, 0x81, 0x80, 0x31, 0x84, 0x25, 0xdf, 0x31, 0xe7, 0x3a, 0xf5, 0x5c,
  0x82, 0xb8, 0x4b, 0x4e, 0x8b, 0x8d, 0xe8, 0x06, 0xa7, 0x3a, 0xa9, 0xb0,
  0x7c, 0xfe, 0x60, 0x3a, 0xca, 0x24, 0x27, 0xae, 0xe5, 0x96, 0x4e, 0x78,
  0xaf, 0x8c, 0x82, 0xb5, 0xfb, 0x0e, 0x5e, 0x73, 0x7c, 0xcc, 0xa9, 0xc3,
  0x47, 0x16, 0x68, 0x78, 0x74, 0x1b, 0x67, 0x5f, 0xc9, 0x30, 0xd3, 0xa1,
  0x59, 0xa2, 0x57, 0x72, 0x44, 0x6e, 0x5e, 0x8b, 0x29, 0x11, 0xc2, 0xd2,
  0x64, 0xc0, 0x31, 0x99, 0x63, 0xd5, 0xbf, 0x3e, 0x71, 0x74, 0xc3, 0x09,
  0x96, 0x5d, 0x4c, 0xf5, 0x42, 0x84, 0x27, 0x00, 0x05, 0x26, 0xda, 0xb7,
  0x02, 0x13, 0x0e, 0xe3, 0x06, 0xa9, 0x7f, 0x0d, 0x58, 0x1b, 0x1c, 0x3d,
  0xf8, 0x36, 0x03, 0x78, 0xb1, 0xd6, 0xab, 0xdc, 0x9f, 0x66, 0xb1, 0x15,
  0xfa, 0x10, 0x0d, 0x0b, 0x8c, 0x4a, 0x00, 0xde, 0xe5, 0x9a, 0xfd, 0x02,
  0x41, 0x00, 0xf7, 0x2d, 0xce, 0xf0, 0x27, 0x50, 0x00, 0xf0, 0x38, 0x96,
  0xc3, 0x01, 0x37, 0xef, 0x97, 0xf6, 0x40, 0x89, 0x47, 0x12, 0xd1, 0x12,
  0x7e, 0xce, 0x42, 0xc4, 0x02, 0xd5, 0x56, 0x20, 0x41, 0x2f, 0x4f, 0x01,
  0x1a, 0x07, 0x55, 0x34, 0xa3, 0xff, 0x3b, 0xf6, 0x5d, 0xd1, 0x10, 0x57,
  0x09, 0x3e, 0xc0, 0x03, 0x05, 0x32, 0x5f, 0xf6, 0x14, 0x3e, 0x6d, 0xea,
  0x9b, 0xd9, 0x5e, 0x6c, 0x12, 0x6f, 0x02, 0x41, 0x00, 0xe7, 0x3b, 0x50,
  0x27, 0xd9, 0x71, 0xfd, 0x63, 0x85, 0x98, 0x56, 0xa9, 0xd2, 0xe2, 0x2e,
  0x6d, 0x50, 0x95, 0x75, 0x85, 0xc9, 0x4f, 0xa3, 0x6a, 0xfd, 0x39, 0x33,
  0x2e, 0x0a, 0xa9, 0x7b, 0xd4, 0x25, 0x11, 0xff, 0x95, 0x31, 0xa6, 0x01,
  0x60, 0xf4, 0x03, 0x98, 0x9b, 0x1e, 0xd9, 0xf4, 0x99, 0x17, 0x37, 0x52,
  0xff, 0xbc, 0xd9, 0x68, 0x17, 0x00, 0x5d, 0xac, 0xb9, 0x42, 0x65, 0xf0,
  0xff, 0x02, 0x41, 0x00, 0xb1, 0xad, 0x0e, 0xf4, 0x97, 0xfd, 0xad, 0x3f,
  0x00, 0x1e, 0x76, 0xfc, 0x38, 0x3e, 0x14, 0xa8, 0xe1, 0x89, 0x0c, 0x3a,
  0xdd, 0x68, 0x83, 0x18, 0x36, 0xeb, 0xf8, 0xf1, 0x6b, 0x20, 0x72, 0x5a,
  0xd4, 0x5a, 0x81, 0x60, 0x46, 0x02, 0x39, 0xe4, 0x6d, 0x30, 0xd2, 0x0d,
  0x46, 0xc6, 0x99, 0x91, 0xae, 0xaa, 0x3d, 0x21, 0x11, 0xd3, 0xac, 0x87,
  0x97, 0xb1, 0x2d, 0xe8, 0x19, 0x34, 0x6c, 0x27, 0x02, 0x40, 0x77, 0xab,
  0xec, 0xda, 0x12, 0xe1, 0xb9, 0xe0, 0x4d, 0xfd, 0xff, 0xac, 0x38, 0xa6,
  0x32, 0x9d, 0x04, 0xbd, 0xb5, 0x6b, 0x1e, 0x3b, 0x07, 0x3a, 0x91, 0xff,
  0x8e, 0x8b, 0x5f, 0x9d, 0x43, 0xa7, 0x11, 0xe0, 0x74, 0xc4, 0x44, 0x43,
  0xc4, 0x26, 0xe7, 0x63, 0x0d, 0x12, 0x0b, 0x2f, 0x86, 0x35, 0xcd, 0x1c,
  0x26, 0xea, 0x90, 0x54, 0x7c, 0x58, 0x9b, 0x29, 0xe0, 0x00, 0xa9, 0x6a,
  0x95, 0x05, 0x02, 0x41, 0x00, 0x90, 0xfb, 0x25, 0xe0, 0xdd, 0xe7, 0x15,
  0x8f, 0x79, 0xde, 0xf0, 0x7c, 0xa4, 0x3d, 0x01, 0xb3, 0x3c, 0xa1, 0x93,
  0x43, 0x00, 0x10, 0xe9, 0x94, 0x11, 0xf7, 0x81, 0xb1, 0xcd, 0xed, 0xeb,
  0xbb, 0xf7, 0x64, 0x3c, 0xb9, 0x04, 0xa9, 0x40, 0xed, 0x90, 0xca, 0xf0,
  0xac, 0x4f, 0x74, 0xe3, 0x5d, 0x5a, 0xea, 0x12, 0x74, 0xec, 0x98, 0x38,
  0x8c, 0x10, 0x0c, 0xf4, 0xfb, 0x55, 0xc0, 0x4c, 0xc6
};


// Username/Passwords Change them!
const char* www_username = "tiberium";
const char* www_password = "tiberium";
const char* ota_password = "tiberium";


// Webserver
#ifdef SECURE
ESP8266WebServerSecure server(443);
#else
ESP8266WebServer server(80);
#endif

char s_amps[8], s_watts[8], s_event[50], s_state[50], s_datetime[50], s_uptime[50];

unsigned int mVperAmp = 100; // use 185 for 10A, 100 for 20A Module and 66 for 30A Module
unsigned int Line = 230;     // Line voltage 230V EU
unsigned int Power = 0;
unsigned int PowerAVG = 0;
float Voltage = 0;
float VRMS = 0;
float AmpsRMS = 0;

//Hold timer
bool on_hold=false;
uint32_t curr_time;
uint32_t last_event;
uint32_t hold_time=10000; // 10 seconds

String ip;

unsigned int relay_state=4; // MANUAL OFF
unsigned int initctype=5;   // LED OFF

// Command Handling
char Command_PC[64], Command[64];
static int Count_PC = 0;
static bool Complete_PC = false;

unsigned long last_wifi_check_time = 0;

char TIBERIUM_Status[200];

/* Pinout */
#define CURRENT_SENSOR A0
#define POWER_RELAY 5   // Fix by relay shield (D1)

// LED pins
int PinRED = 0;   
int PinGREEN = 14;
int PinBLUE = 16;

// Button
int ButtonPin = 12;

// Smoothing
const int numReadings = 200;
int readings[numReadings];      // the readings from the analog input
int readIndex = 0;              // the index of the current reading
unsigned long total = 0;        // the running total
int average = 0;                // the average

/* SPIFFS Flash Storage */
#include "FS.h"
#define DATAFILE1 "/tiberium-state.db"
#define DATAFILE2 "/tiberium-time.db"
#define DATAFILE3 "/timer-state.db"

bool spiffsActive = false;
void spiffs_start() {
    if (SPIFFS.begin()) {
      Serial.println("SPIFFS Active");
      Serial.println();
      spiffsActive = true;
  } else {
      Serial.println("Unable to activate SPIFFS");
  }
}

void spiffs_readdata() {
    if (spiffsActive) {
    if (SPIFFS.exists(DATAFILE1)) {
      File f = SPIFFS.open(DATAFILE1, "r");
      if (!f) {
        #ifdef DEBUG
        Serial.print("Unable To Open '");
        Serial.print(DATAFILE1);
        Serial.println("' for Reading");
        #endif
      } else {
        char data; int rb_state;
        #ifdef DEBUG
        Serial.print("Contents of file '");
        Serial.print(DATAFILE1);
        Serial.println("'");
        Serial.println();
        #endif
        while (f.position()<f.size())
        {
          data = char (f.read());               
        } 
        
        rb_state=(int)data-48;

        #ifdef DEBUG
        Serial.print("Last state:");
        Serial.println(rb_state);
        #endif

        if (rb_state<=3)
        {
         relay_on(rb_state);
        }        
        else if (rb_state>3)
        {
         relay_off(rb_state-3);
        }

        f.close();
      }
    }
   }
}

void spiffs_writedata() {
    if (spiffsActive) {    
      File f = SPIFFS.open(DATAFILE1, "w");
      if (!f) {
        #ifdef DEBUG
        Serial.print("Unable To Open '");
        Serial.print(DATAFILE1);
        Serial.println("' for Writing");
        #endif
      } else {
        f.print(relay_state);
        f.close();
      }    
   }
}

/* TIMER */

// Update globs, returns nothing 
void tme_readdata() {
    if (spiffsActive) {
    if (SPIFFS.exists(DATAFILE3)) {
      File f = SPIFFS.open(DATAFILE3, "r");
      if (!f) {
        #ifdef DEBUG
        Serial.print("Unable To Open '");
        Serial.print(DATAFILE3);
        Serial.println("' for Reading");
        #endif
      } else {
        char data; int timer_state;
        #ifdef DEBUG
        Serial.print("Contents of file '");
        Serial.print(DATAFILE3);
        Serial.println("'");
        Serial.println();
        #endif
        while (f.position()<f.size())
        {
          data = char (f.read());               
        } 
        
        timer_state=(int)data-48;

        #ifdef DEBUG
        Serial.print("Timer Last state:");
        Serial.println(timer_state);
        #endif

        if (timer_state==0)
        {
         timer_enabled=0;
        }        
        else if (timer_state==1)
        {
         timer_enabled=1;
        }

        f.close();
      }
    }
   }
}

// Update datafile, returns nothing
void tme_writedata() {
    if (spiffsActive) {    
      File f = SPIFFS.open(DATAFILE3, "w");
      if (!f) {
        #ifdef DEBUG
        Serial.print("Unable To Open '");
        Serial.print(DATAFILE3);
        Serial.println("' for Writing");
        #endif
      } else {
        f.print(timer_enabled);
        f.close();
      }    
   }
}

// Returns 1 on ok 0 otherwise (time DB has to have exactly 10 entries)
int timer_vrfydb() {
  unsigned int i=0;
  char cr;
  
  File f = SPIFFS.open(DATAFILE2, "r");
     if (!f) {
       return 0;
      } else {

       while (f.position()<f.size())
        {
          cr = f.read();               
          if(cr == '\n')
           i++;
        }
       f.close(); 
      }      
        
   if (i != 10)
    return 0;
   else
    return 1;
}

/* Wish if there were a better way */
void TimeFormGenerator(unsigned int time_entry)
{
  unsigned int t_type=0, t_hh=0, t_mm=0;
  String t_name;
  
  if(time_entry < 1 || time_entry > 10)
   return;
   
      switch (time_entry) {
        case 1:
        t_type=t1_set;
        t_hh=t1_hh;
        t_mm=t1_mm;
        break;
        case 2:
        t_type=t2_set;
        t_hh=t2_hh;
        t_mm=t2_mm;
        break;
        case 3:
        t_type=t3_set;
        t_hh=t3_hh;
        t_mm=t3_mm;
        break;
        case 4:
        t_type=t4_set;
        t_hh=t4_hh;
        t_mm=t4_mm;
        break;
        case 5:
        t_type=t5_set;
        t_hh=t5_hh;
        t_mm=t5_mm;
        break;
        case 6:
        t_type=t6_set;
        t_hh=t6_hh;
        t_mm=t6_mm;
        break;
        case 7:
        t_type=t7_set;
        t_hh=t7_hh;
        t_mm=t7_mm;
        break;
        case 8:
        t_type=t8_set;
        t_hh=t8_hh;
        t_mm=t8_mm;
        break;
        case 9:
        t_type=t9_set;
        t_hh=t9_hh;
        t_mm=t9_mm;
        break;
        case 10:
        t_type=t10_set;
        t_hh=t10_hh;
        t_mm=t10_mm;
        break;        
       }

if(t_type==0) t_name="DISABLED";
if(t_type==1) t_name="ON";
if(t_type==2) t_name="OFF";

String response="<tr><td>Timer " + String(time_entry) +":</td><td><FORM action=\"/timer\" method=\"get\">"
"<input type=\"hidden\" id=\"entry\" name=\"entry\" value=\""+ String(time_entry) +"\">"
"<select name=\"type\">"
"<option selected=\""+ String(t_type) +"\">"+ t_name +"</option>"
"<option value=\"0\">DISABLED</option>"
"<option value=\"1\">ON</option>"
"<option value=\"2\">OFF</option>"
"</select>";
server.sendContent(response);

server.sendContent("<select name=\"hh\">"
"<option selected=\""+ String(t_hh) +"\">"+ String(t_hh) +"</option>"
"<option value=\"0\">00</option>"
"<option value=\"1\">01</option>"
"<option value=\"2\">02</option>"
"<option value=\"3\">03</option>"
"<option value=\"4\">04</option>"
"<option value=\"5\">05</option>"
"<option value=\"6\">06</option>"
"<option value=\"7\">07</option>"
"<option value=\"8\">08</option>"
"<option value=\"9\">09</option>"
"<option value=\"10\">10</option>"
"<option value=\"11\">11</option>"
"<option value=\"12\">12</option>"
"<option value=\"13\">13</option>"
"<option value=\"14\">14</option>"
"<option value=\"15\">15</option>"
"<option value=\"16\">16</option>"
"<option value=\"17\">17</option>"
"<option value=\"18\">18</option>"
"<option value=\"19\">19</option>"
"<option value=\"20\">20</option>"
"<option value=\"21\">21</option>"
"<option value=\"22\">22</option>"
"<option value=\"23\">23</option>"
"</select>");

server.sendContent("<select name=\"mm\">"
"<option selected=\""+ String(t_mm) +"\">"+ String(t_mm) +"</option>"
"<option value=\"0\">00</option>"
"<option value=\"1\">01</option>"
"<option value=\"2\">02</option>"
"<option value=\"3\">03</option>"
"<option value=\"4\">04</option>"
"<option value=\"5\">05</option>"
"<option value=\"6\">06</option>"
"<option value=\"7\">07</option>"
"<option value=\"8\">08</option>"
"<option value=\"9\">09</option>"
"<option value=\"10\">10</option>"
"<option value=\"11\">11</option>"
"<option value=\"12\">12</option>"
"<option value=\"13\">13</option>"
"<option value=\"14\">14</option>"
"<option value=\"15\">15</option>"
"<option value=\"16\">16</option>"
"<option value=\"17\">17</option>"
"<option value=\"18\">18</option>"
"<option value=\"19\">19</option>"
"<option value=\"20\">20</option>"
"<option value=\"21\">21</option>"
"<option value=\"22\">22</option>"
"<option value=\"23\">23</option>"
"<option value=\"24\">24</option>"
"<option value=\"25\">25</option>"
"<option value=\"26\">26</option>"
"<option value=\"27\">27</option>"
"<option value=\"28\">28</option>"
"<option value=\"29\">29</option>");

server.sendContent("<option value=\"30\">30</option>"
"<option value=\"31\">31</option>"
"<option value=\"32\">32</option>"
"<option value=\"33\">33</option>"
"<option value=\"34\">34</option>"
"<option value=\"35\">35</option>"
"<option value=\"36\">36</option>"
"<option value=\"37\">37</option>"
"<option value=\"38\">38</option>"
"<option value=\"39\">39</option>"
"<option value=\"40\">40</option>"
"<option value=\"41\">41</option>"
"<option value=\"42\">42</option>"
"<option value=\"43\">43</option>"
"<option value=\"44\">44</option>"
"<option value=\"45\">45</option>"
"<option value=\"46\">46</option>"
"<option value=\"47\">47</option>"
"<option value=\"48\">48</option>"
"<option value=\"49\">49</option>"
"<option value=\"50\">50</option>"
"<option value=\"51\">51</option>"
"<option value=\"52\">52</option>"
"<option value=\"53\">53</option>"
"<option value=\"54\">54</option>"
"<option value=\"55\">55</option>"
"<option value=\"56\">56</option>"
"<option value=\"57\">57</option>"
"<option value=\"58\">58</option>"
"<option value=\"59\">59</option>"
"</select>"
"<INPUT type=\"submit\" value=\"SET\"></form></td></tr>");

 return;
}

// Only updates Globs, does not touch the DB
void SetTimeVar(unsigned int time_var, unsigned int t_type, unsigned int t_hh, unsigned int t_mm)
{
      switch (time_var) {
        case 1:
        t1_set=t_type;
        t1_hh=t_hh;
        t1_mm=t_mm;
        break;
        case 2:
        t2_set=t_type;
        t2_hh=t_hh;
        t2_mm=t_mm;
        break;
        case 3:
        t3_set=t_type;
        t3_hh=t_hh;
        t3_mm=t_mm;
        break;
        case 4:
        t4_set=t_type;
        t4_hh=t_hh;
        t4_mm=t_mm;
        break;
        case 5:
        t5_set=t_type;
        t5_hh=t_hh;
        t5_mm=t_mm;
        break;
        case 6:
        t6_set=t_type;
        t6_hh=t_hh;
        t6_mm=t_mm;
        break;
        case 7:
        t7_set=t_type;
        t7_hh=t_hh;
        t7_mm=t_mm;
        break;
        case 8:
        t8_set=t_type;
        t8_hh=t_hh;
        t8_mm=t_mm;
        break;
        case 9:
        t9_set=t_type;
        t9_hh=t_hh;
        t9_mm=t_mm;
        break;
        case 10:
        t10_set=t_type;
        t10_hh=t_hh;
        t10_mm=t_mm;
        break;
        
       }
}

/* This function reads data from the DB file and set it in global vars.
 * This function should be only called ONCE at start to read the DB. 
 * All the other operations like data query should be done from the global vars.
 * 
 *  Each row in the DB corresponds to it's tX entry.
 *  The first column: type: 0 - timer entry disabled | 1 - timer ON event | 2 - timer OFF event
 *  Second column: hours
 *  Third column: minutes
 */
void ReadLineNToGlob(unsigned int lineNumber) {
  char data_column1[5], data_column2[5], data_column3[5], time_string[50];
  bool EOL=false;  
  char cr;

/* Zero out these otherwise... */
  unsigned int p=0, q=0, r=0, cc=0;
  unsigned int t_type=0, t_hh=0, t_mm=0;

 memset(data_column1, 0, sizeof(data_column1));
 memset(data_column2, 0, sizeof(data_column2));
 memset(data_column3, 0, sizeof(data_column3));
 memset(time_string, 0, sizeof(time_string));
 
  if (lineNumber <1 || lineNumber > 11) {
    #ifdef DEBUG
     Serial.println("I don't want to segfault :(");
    #endif 
    return;
  }
  
  File f = SPIFFS.open(DATAFILE2, "r");
     if (!f) {
        #ifdef DEBUG
        Serial.print("Unable To Open '");
        Serial.print(DATAFILE2);
        Serial.println("' for reading");
        #endif
      } else {
        f.seek(0);

  // Seek to the end of the previous line
        for(unsigned int i = 0; i < (lineNumber -1);){
         cr = f.read();
          if(cr == '\n'){
          i++;
         }
        }
  
  // Now we are at the right line YaY
  while(!EOL){
    cr = f.read(); // Read line char by char

    if (cr == ',') 
       {
        cc++;
       }
       else if (cr != '\n')
       {
         if(cc==0)
         data_column1[p++]= cr;
         
         if(cc==1)
         data_column2[q++]= cr;
         
         if(cc==2)
         data_column3[r++]= cr;
       }
     
    if(cr == '\n')
      EOL=true;
  }
  
  f.close();

  // Line assembled
  t_type=(unsigned int)atoi(data_column1);
  t_hh=(unsigned int)atoi(data_column2);
  t_mm=(unsigned int)atoi(data_column3);
  
    #ifdef DEBUG
      Serial.print("DC1: ");
      Serial.println(t_type);
      Serial.print("DC2: ");
      Serial.println(t_hh);
      Serial.print("DC3: ");
      Serial.println(t_mm);
      //Serial.print("TS: ");
      //Serial.println(time_string);
    #endif

/* FS -> global var only */
    SetTimeVar(lineNumber, t_type, t_hh,  t_mm);
       
  }
}

void timer_readdb() {
  #ifdef DEBUG
   Serial.println("Timer DB");
   Serial.println("--------");
  #endif
  for (unsigned int x=1;x<11;x++)  // There is a reason why this goes from 1-11 instead of 0-10
   {
    #ifdef DEBUG
     Serial.print("Entry : ");
     Serial.println(x);
    #endif
     ReadLineNToGlob(x);
   }
}

//In case of ANY modifications dump the whole time DB from env vars and write it out to the file (this should rarely happen)
void timer_writedb(bool initdb) {
    char time_string[50];
    if (spiffsActive) {    
      File f = SPIFFS.open(DATAFILE2, "w");
      if (!f) {
        #ifdef DEBUG
        Serial.print("Unable To Open '");
        Serial.print(DATAFILE2);
        Serial.println("' for Writing");
        #endif
      } else {

      if (initdb) {
        for (unsigned int i=1;i<=10;i++)
        {
         sprintf(time_string,"%d,%d,%d",0,00,00);
         f.println(time_string);
        }
      } else {      
         sprintf(time_string,"%d,%d,%d",t1_set,t1_hh,t1_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t2_set,t2_hh,t2_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t3_set,t3_hh,t3_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t4_set,t4_hh,t4_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t5_set,t5_hh,t5_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t6_set,t6_hh,t6_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t7_set,t7_hh,t7_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t8_set,t8_hh,t8_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t9_set,t9_hh,t9_mm);
         f.println(time_string);
         sprintf(time_string,"%d,%d,%d",t10_set,t10_hh,t10_mm);
         f.println(time_string);
      }
         f.close();
      }    
   }
}   

void handleTimer() {
 unsigned int entry=0, type=0, hh=0, mm=0;
  
 entry = (unsigned int)server.arg(0).toInt();
 type = (unsigned int)server.arg(1).toInt();
 hh = (unsigned int)server.arg(2).toInt();
 mm = (unsigned int)server.arg(3).toInt();

// Do some input validation!
// Hours never reach 24 as minutes never reach 60
if (server.args() != 4 || entry < 1 || entry > 10 || type < 0 || type > 3 || hh > 23 || hh < 0 || mm < 0 || mm > 59 )
 {
   server.send(200, "text/html", "<!doctype html><html><body><center>Invalid request.</center></body></html>");
   return;
 }
   
#ifdef DEBUG
Serial.print("Entry: ");
Serial.print(entry);
Serial.print(" Type: ");
Serial.print(type);
Serial.print(" HH: ");
Serial.print(hh);
Serial.print(" MM: ");
Serial.println(mm);
/*
String message = "Number of args received:";
message += server.args();
message += "\n";         
for (int i = 0; i < server.args(); i++) {
 message += "Arg no" + (String)i + " –> ";   
 message += server.argName(i) + ": ";    
 message += server.arg(i) + "\n";        
}
Serial.println(message);
*/
#endif
 SetTimeVar(entry, type, hh,  mm); // Update globs
 timer_writedb(0); // Write out DB

 server.send(200, "text/html", "<!doctype html><html><body><script type=\"text/javascript\">location.href = '/';</script><center><a href=\"/\">Time saved.</a></center></body></html>");
} 

/* Timer Enabler */
void handleTme() {
 unsigned int timer=0;

 timer = (unsigned int)server.arg(0).toInt();

// Do some input validation!
if (server.args() != 1 || timer > 1 || timer < 0 )
 {
   server.send(200, "text/html", "<!doctype html><html><body><center>Invalid request.</center></body></html>");
   return;
 }
   
#ifdef DEBUG
 Serial.print("Timer State: ");
 Serial.print(timer);
#endif

 timer_enabled=timer; // Update globs
 tme_writedata(); // Write out DB

 server.send(200, "text/html", "<!doctype html><html><body><script type=\"text/javascript\">location.href = '/';</script><center><a href=\"/\">Timer enabled.</a></center></body></html>");
} 

/* Timer Clear */
void tmc() {
 // Disable the timer itself
 timer_enabled=0; // Update globs
 tme_writedata(); // Write out DB

 for (unsigned int i=1; i<=10;i++) // Reset global variables
      SetTimeVar(i, 0, 0, 0);

 timer_writedb(1); // Reset the DB on flash
}

void handleTmc() {
 unsigned int tc=0;

 tc = (unsigned int)server.arg(0).toInt();

// Do some input validation!
if (server.args() != 1 || tc != 1 )
 {
   server.send(200, "text/html", "<!doctype html><html><body><center>Invalid request.</center></body></html>");
   return;
 }
   
#ifdef DEBUG
 Serial.print("Timer Clear: ");
 Serial.print(tc);
#endif

 tmc();
 
 server.send(200, "text/html", "<!doctype html><html><body><script type=\"text/javascript\">location.href = '/';</script><center><a href=\"/\">Timer reset.</a></center></body></html>");
} 

/* Wifi Connect */
void wifi_connect() {

 unsigned int conncnt=0;

#ifdef DEBUG	
  Serial.print("Connecting to ");
  Serial.println(ssid);
#endif  

  WiFi.hostname(NODENAME);
  WiFi.setSleepMode(WIFI_NONE_SLEEP);  
  WiFi.begin(ssid, pass);
#ifdef STATIC_IP  
  WiFi.config(staticIP, dns, gateway, subnet);  // MUST be after Wifi begin how dumb... -> this is the ESP8266 order, order for the ESP32 is different!
#endif
  WiFi.mode(WIFI_STA); // MUST otherwise it will also create an unwanted AP

/* This check will pass no matter what if staic ip is used */
  while (WiFi.status() != WL_CONNECTED) {
    ledcolor(4); // No need for delay this is blocking
    delay(200);  // Extra delay added for stability
#ifdef DEBUG
    Serial.print(".");
#endif  
  conncnt++;

/* If it can't connect to the wifi for a while restart. */  
  if(conncnt==30)
    ESP.restart();

  }

/* Restores led color to whatever was read back from the db initially to clear out the yellow blinking. */
  ledcolor(initctype);
  
ip=WiFi.localIP().toString();

#ifdef DEBUG  
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(ip); 
#endif  
}

/* 1 - manual, 2 - time , 3 - computer control */
void relay_on(unsigned int control_type) { 
 digitalWrite(POWER_RELAY, HIGH);
 char stat[20];
  switch (control_type) {
  case 1:  // red
   ledcolor(control_type);
   strcpy(stat,"MANUAL");
  break;
  case 2: // green
   ledcolor(control_type);
   strcpy(stat,"TIME");
  break;
  case 3: // blue
   ledcolor(control_type);
   strcpy(stat,"COMP");
  break;  
}

  initctype=control_type;
  
  sprintf(s_event,"%s_%s%s", NODENAME, stat, "_ON_OK");
  sprintf(s_state,"%s_%s%s", NODENAME, stat, "_ON");

 #ifdef DEBUG
  Serial.println(s_event);
 #endif
 
 relay_state=control_type;
 spiffs_writedata(); // Write to flash
}

/* 
   1 - manual, 2 - time , 3 - computer control 
   will result relay_state
   4 - manual_off, 5 - time_off, 6 - computer_off
*/
void relay_off(unsigned int control_type) {
 digitalWrite(POWER_RELAY, LOW);
 char stat[20]; 
  switch (control_type) {
  case 1:  // off
   ledcolor(6);
   strcpy(stat,"MANUAL");
  break;
  case 2: // off
   ledcolor(6);
   strcpy(stat,"TIME");
  break;
  case 3: // off
   ledcolor(6);
   strcpy(stat,"COMP");
  break;  
} 

  sprintf(s_event,"%s_%s%s", NODENAME, stat, "_OFF_OK");
  sprintf(s_state,"%s_%s%s", NODENAME, stat, "_OFF");

 #ifdef DEBUG
  Serial.println(s_event);
 #endif

 relay_state=control_type+3;
 spiffs_writedata(); // Write to flash
}


void ledcolor(unsigned int color) {

switch (color) {
  case 1:  // red
  //Serial.println("red");
   digitalWrite(PinRED, LOW);
   digitalWrite(PinGREEN, HIGH);
   digitalWrite(PinBLUE, HIGH);
  break;
  case 2: // green
  //Serial.println("green");
   digitalWrite(PinRED, HIGH);
   digitalWrite(PinGREEN, LOW);
   digitalWrite(PinBLUE, HIGH);  
  break;
  case 3: // blue
  //Serial.println("blue");
   digitalWrite(PinRED, HIGH);
   digitalWrite(PinGREEN, HIGH);
   digitalWrite(PinBLUE, LOW);
  break;
  case 4: // yellow blinking
  //Serial.println("yellow");
   digitalWrite(PinRED, LOW);
   digitalWrite(PinGREEN, LOW);
   digitalWrite(PinBLUE, HIGH);  
   delay(500);
   ledcolor(6);
   delay(500);
  break;
  case 5: // purple blinking
  //Serial.println("purple");
   digitalWrite(PinRED, LOW);
   digitalWrite(PinGREEN, HIGH);
   digitalWrite(PinBLUE, LOW);  
   delay(500);
   ledcolor(6);
   delay(500);
  break;
  case 6: // off
  //Serial.println("off");
   digitalWrite(PinRED, HIGH);
   digitalWrite(PinGREEN, HIGH);
   digitalWrite(PinBLUE, HIGH);  
  break;
  }
  
}

void update_status() {
  timeClient.update();
  timeClient.getFormattedDate(0,0).toCharArray(s_datetime,50);
  timeClient.getFormattedDate(1,0).toCharArray(s_uptime,50);
  sprintf(TIBERIUM_Status, "%s,%s,%s,%s,%d,%d", s_state, s_datetime, s_uptime, s_amps, Power, PowerAVG);
  #ifdef DEBUG
   Serial.println(TIBERIUM_Status);
  #endif
}

String getValue(String data, char separator, int index)
{
    int found = 0;
    int strIndex[] = { 0, -1 };
    int maxIndex = data.length() - 1;

    for (int i = 0; i <= maxIndex && found <= index; i++) {
        if (data.charAt(i) == separator || i == maxIndex) {
            found++;
            strIndex[0] = strIndex[1] + 1;
            strIndex[1] = (i == maxIndex) ? i+1 : i;
        }
    }
    return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}

/*
 Network Services Initialization
*/
void net_init() {
  // NTP initial sync   
 timeClient.begin();
 timeClient.update();

/* Don't proceed if the date is invalid, try it forever */
while (timeClient.getFormattedDate(0,1).toInt() < 2020) {

 timeClient.update();
 ledcolor(5);
#ifdef DEBUG
 Serial.print("This year is: ");
 Serial.println(timeClient.getFormattedDate(0,1).toInt());
#endif
  delay(100);
}

/* Restore led color to whatever was read back from the db initially to clear out the yellow blinking. */
 ledcolor(initctype);

 timeClient.setStartEpochTime();

// Global status update
 update_status();
   
#ifdef SECURE
// Webserver (HTTPS)
 server.getServer().setServerKeyAndCert_P(rsakey, sizeof(rsakey), x509, sizeof(x509));
#endif

/* These are the ONLY possible URIs everything else will result 404 */
 server.on("/", handleRoot);  
 server.on("/on", handleRoot);  
 server.on("/off", handleRoot);  
 server.on("/restart", handleRoot);
 server.on("/m2m", handleRoot);  
 server.on("/m2m/", handleRoot);  
 server.on("/m2m/on", handleRoot);  
 server.on("/m2m/off", handleRoot);  

/* Server for the timer */
 server.on("/timer", handleTimer);  
 server.on("/tme", handleTme);  
 server.on("/tmc", handleTmc);  

 // Start the server
 server.begin();

 
 for (int thisReading = 0; thisReading < numReadings; thisReading++) {
    readings[thisReading] = 0;
  }

// OTA
#ifdef OTA
  ArduinoOTA.setPassword(ota_password);
  ArduinoOTA.onStart([]() {
    String type;
    if (ArduinoOTA.getCommand() == U_FLASH) {
      type = "sketch";
    } else { // U_FS
      type = "filesystem";
    }
#ifdef DEBUG
    // NOTE: if updating FS this would be the place to unmount FS using FS.end()
    Serial.println("Start updating " + type);
#endif
  });


#ifdef DEBUG  
  ArduinoOTA.onEnd([]() {
    Serial.println("\nEnd");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) {
      Serial.println("Auth Failed");
    } else if (error == OTA_BEGIN_ERROR) {
      Serial.println("Begin Failed");
    } else if (error == OTA_CONNECT_ERROR) {
      Serial.println("Connect Failed");
    } else if (error == OTA_RECEIVE_ERROR) {
      Serial.println("Receive Failed");
    } else if (error == OTA_END_ERROR) {
      Serial.println("End Failed");
    }
  });
#endif
  
  ArduinoOTA.begin();
#endif
// OTA
}

void setup() {

 pinMode(ButtonPin,INPUT); 
 pinMode(POWER_RELAY, OUTPUT);
 pinMode(PinRED, OUTPUT);
 pinMode(PinGREEN, OUTPUT);
 pinMode(PinBLUE, OUTPUT);
  
 
#ifdef DEBUG
 Serial.begin(115200);  // USB
#endif

/* 
 Sample Status:
 %Node Name_%Control Type%_ON/OFF%,%DATETIME%,%UPTIME%,%Current%,%Power%,%PowerAVG% 
*/
 strcpy(s_state, NODENAME"_INIT");
 strcpy(TIBERIUM_Status, NODENAME"_INIT,0000-00-00T00:00:00,0000-00-00T00:00:00,0,0,0");          

// Flash storage
 spiffs_start();
// Settings readback from flash storage
 spiffs_readdata(); // if the file can't be opened it defaults to 4 which is MANUAL OFF

// Time switch
 tme_readdata(); // First check if timer is enabled (this has no effect on the existing timer DB, if the file can't be opened it won't be created it just defaults to 0 which is DISABLED

 if (!timer_vrfydb())
 {
   #ifdef DEBUG
    Serial.println("Timer db does not exist or corrupted, recreating...");
   #endif
    timer_writedb(true);
 } 
 timer_readdb(); // Load timer events from DB -> Global variables

// Wifi connect
 wifi_connect();

// Network services initialization
 net_init();  
}

  
/* Incoming command handling */

#ifdef DEBUG
void xcomm() {
/* COMM FROM PC */   
if (Serial.available()){

    Serial.println("PC: New command, collecting...");
    memset(Command_PC, 0, sizeof(Command_PC));
    while (Serial.available()){
      char character = Serial.read();
      if (character=='\n')
         {
          Complete_PC =true;
          Count_PC = 0;
         } else {
             if (Count_PC < 62) { // Don't overflow the array...!
                 Command_PC[Count_PC++] = character;
                 Command_PC[Count_PC] = 0;
             }
         }
  }
}   

if (Complete_PC) {
  
    strcpy(Command, Command_PC);
    Serial.print("Command received (PC): >");
    Serial.print(Command_PC);
    Serial.println("<");

    if (strcmp(Command, NODENAME"_ON")  == 0) {
              relay_on(3);
            }             
    else if (strcmp(Command, NODENAME"_OFF")  == 0) {
              relay_off(3);
    }
    else if (strcmp(Command, "TIME_CLEAR")  == 0) {
              tmc();
              Serial.println("Timer DB cleared.");                     
            }                       
    else if (strcmp(Command, "TIME_QUERY")  == 0) {
              timer_readdb();
            }                       
    else if (strcmp(Command, "TBR_STAT")  == 0) {
              Serial.println(TIBERIUM_Status);                     
            }                       
           Complete_PC=false;      
 }
}
#endif

/* COMM FROM WEB */   
void handleRoot() {
 
            String response;

#ifdef DEBUG            
    Serial.println("Client connected.");
#endif


            // Auth
            if (!server.authenticate(www_username, www_password))
              return server.requestAuthentication();

            // Process the command right away
            if (server.uri() == "/m2m/on") {
               relay_on(3);               
               server.send(200, "text/plain", NODENAME"_ON_OK\n");    
            } else if (server.uri() == "/m2m/off") {
              relay_off(3);
              server.send(200, "text/plain", NODENAME"_OFF_OK\n");    
            } else if (server.uri() == "/restart") {
              ESP.restart();
            } else if (server.uri() == "/m2m" || server.uri() == "/m2m/") {
               char tbr[100];
               sprintf(tbr, "%s\n", TIBERIUM_Status);
               server.send(200, "text/plain", tbr);    
            } else if (server.uri() == "/" || server.uri() == "/on" || server.uri() == "/off") {

// These are NOT manual ON/OFFs, it's used by the webpage so it has to fall through for the page generation
            if (server.uri() == "/on")
              relay_on(3);

            if (server.uri() == "/off")
              relay_off(3);

             server.setContentLength(CONTENT_LENGTH_UNKNOWN);

            // Top
            
            response += "<!DOCTYPE html><html><head><meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">";
            response += "<link rel=\"icon\" href=\"data:,\"><meta charset=\"utf-8\"/><title>"NODENAME"</title>";
            response += "<style>html { font-family: Helvetica; display: inline-block; margin: 0px auto; text-align: center;}";
            response += ".button { background-color: #F44336; border: none; color: white; padding: 16px 40px;";
            response += "text-decoration: none; font-size: 30px; margin: 2px; cursor: pointer;}";
            response += ".button2 {background-color: #008CBA;}</style></head>";                        
            response += "<body><center><table><tr><td><h1>Node:</h1></td><td><h1><font color=\"#0000FF\"> "NODENAME"</font></h1></td></tr>";

            server.send(200, "text/html", response);

            server.sendContent("<tr><td>IP:</td><td>"+ ip +"</td></tr>");

            server.sendContent("<tr><td>Date:</td><td>"+ String(s_datetime) +"</td></tr>");

            server.sendContent("<tr><td>Uptime:</td><td>"+ String(s_uptime) +"</td></tr>");

            server.sendContent("<tr><td>Version:</td><td>"+ String(VERSION) +"</td></tr>");

            server.sendContent("<tr><td>Current [A]:</td><td>"+ String(s_amps) +"</td></tr>");

            server.sendContent("<tr><td>Power [W]:</td><td>"+ String(Power) +"</td></tr>");

            server.sendContent("<tr><td>Power AVG [W]:</td><td>"+ String(PowerAVG) +"</td></tr>");


            if(timer_enabled)
              server.sendContent("<tr><td>Timer:</td><td><FORM action=\"/tme\" method=\"get\"><input type=\"hidden\" name=\"timer\" value=\"0\"><input type=\"submit\" value=\"DISABLE\"></FORM></td></tr>");
            else
              server.sendContent("<tr><td>Timer:</td><td><FORM action=\"/tme\" method=\"get\"><input type=\"hidden\" name=\"timer\" value=\"1\"><input type=\"submit\" value=\"ENABLE\"></FORM></td></tr>");
             
            server.sendContent("<tr><td>Restart plug:</td><td><a href=\"/restart\"><button>RESTART</button></a></td></tr>");
            
            if (relay_state>3)
              server.sendContent("<tr><td>Plug state:</td><td>"+ String (s_state) +"</td></tr><tr><td colspan=\"2\" align=\"center\"><a href=\"/on\"><button class=\"button\">ON</button></a></td></tr>");
            else
              server.sendContent("<tr><td>Plug state:</td><td>"+ String (s_state) +"</td></tr><tr><td colspan=\"2\" align=\"center\"><a href=\"/off\"><button class=\"button button2\">OFF</button></a></td></tr>");

            

            if(timer_enabled)
            {
              server.sendContent("<tr><td>&nbsp;</td><td>&nbsp;</td></tr><tr><td>&nbsp;</td><td>&nbsp;</td></tr>");
              //Time Entries
              for (unsigned int z=1;z<=10;z++)
                  TimeFormGenerator(z);

              server.sendContent("<tr><td>Clear timer db:</td><td><FORM action=\"/tmc\" method=\"get\"><input type=\"hidden\" name=\"tc\" value=\"1\"><input type=\"submit\" value=\"RESET\"></FORM></td></tr>");
            }
            
           //Footer
           server.sendContent("</table></center></body></html>");
           server.sendContent("");
           server.client().stop();
              
            } else {
              server.send(200, "text/plain", "404\n");
            }
#ifdef DEBUG            
    Serial.println("Client disconnected.");
#endif
}

/* AC current monitor. Locks while sampling */
float AC_VPP()
{
  float result;
  
  int readValue;             // value read from the sensor
  int maxValue = 0;          // store max value here
  int minValue = 1024;       // store min value here
  
   uint32_t start_time = millis();
   while((millis()-start_time) < 1000) //sample for 1 Sec
   {
       readValue = analogRead(CURRENT_SENSOR);
       // see if you have a new maxValue
       if (readValue > maxValue) 
       {
           /*record the maximum sensor value*/
           maxValue = readValue;
       }
       if (readValue < minValue) 
       {
           /*record the maximum sensor value*/
           minValue = readValue;
       }
   }
   
   // Subtract min from max
   //result = ((maxValue - minValue) * 5.0)/1024.0;
   result = ((maxValue - minValue) * 3.3)/1024.0;   
   return result;
}
 
void loop() {
    curr_time = millis();

    // Limit events
    if (on_hold && curr_time - last_event >= hold_time) {
        on_hold = false;
    }

// Wifi connection check
  if(curr_time - last_wifi_check_time > WIFI_TIMEOUT) {
#ifdef DEBUG
    Serial.println("Checking WiFi... ");
#endif    
    if(WiFi.status() != WL_CONNECTED) {
#ifdef DEBUG      
      Serial.println("WiFi connection lost. Reconnecting...");
#endif
      wifi_connect(); // This is blocking we will ONLY get to the init section if this succeeds 
    }     
    last_wifi_check_time = curr_time;
  }

// NTP update (this will be run each loop but the event is limited by the function)
update_status();

// TIME Switch
if(timer_enabled)
{
 if (t1_set>0)
  if (t1_hh == timeClient.getHours(0) && t1_mm == timeClient.getMinutes(0))
   {
    if (t1_set==1)
       relay_on(2);
    else if (t1_set==2)
       relay_off(2);
   }
 if (t2_set>0)
  if (t2_hh == timeClient.getHours(0) && t2_mm == timeClient.getMinutes(0))
   {
    if (t2_set==1)
       relay_on(2);
    else if (t2_set==2)
       relay_off(2);
   }
 if (t3_set>0)
  if (t3_hh == timeClient.getHours(0) && t3_mm == timeClient.getMinutes(0))
   {
    if (t3_set==1)
       relay_on(2);
    else if (t3_set==2)
       relay_off(2);
   }
 if (t4_set>0)
  if (t4_hh == timeClient.getHours(0) && t4_mm == timeClient.getMinutes(0))
   {
    if (t4_set==1)
       relay_on(2);
    else if (t4_set==2)
       relay_off(2);
   }
 if (t5_set>0)
  if (t5_hh == timeClient.getHours(0) && t5_mm == timeClient.getMinutes(0))
   {
    if (t5_set==1)
       relay_on(2);
    else if (t5_set==2)
       relay_off(2);
   }
 if (t6_set>0)
  if (t6_hh == timeClient.getHours(0) && t6_mm == timeClient.getMinutes(0))
   {
    if (t6_set==1)
       relay_on(2);
    else if (t6_set==2)
       relay_off(2);
   }
 if (t7_set>0)
  if (t7_hh == timeClient.getHours(0) && t7_mm == timeClient.getMinutes(0))
   {
    if (t7_set==1)
       relay_on(2);
    else if (t7_set==2)
       relay_off(2);
   }
 if (t8_set>0)
  if (t8_hh == timeClient.getHours(0) && t8_mm == timeClient.getMinutes(0))
   {
    if (t8_set==1)
       relay_on(2);
    else if (t8_set==2)
       relay_off(2);
   }
 if (t9_set>0)
  if (t9_hh == timeClient.getHours(0) && t9_mm == timeClient.getMinutes(0))
   {
    if (t9_set==1)
       relay_on(2);
    else if (t9_set==2)
       relay_off(2);
   }
 if (t10_set>0)
  if (t10_hh == timeClient.getHours(0) && t10_mm == timeClient.getMinutes(0))
   {
    if (t10_set==1)
       relay_on(2);
    else if (t10_set==2)
       relay_off(2);
   }
}
  
// Comm code 
#ifdef DEBUG   
 xcomm();
#endif
 
// Push button
 int reading = digitalRead(ButtonPin);

  // If the switch changed, due to noise or pressing:
  if (reading != lastButtonState) {
    // reset the debouncing timer
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) > debounceDelay) {
    // whatever the reading is at, it's been there for longer than the debounce
    // delay, so take it as the actual current state:

    // if the button state has changed:
    if (reading != buttonState) {
      buttonState = reading;

//     Serial.println("Button push detected");
      // only toggle the LED if the new button state is HIGH
      if (buttonState == HIGH && relay_state > 3)
       relay_on(1);
      else if (buttonState == HIGH && relay_state <= 3)
       relay_off(1);
      
    }
  }
  lastButtonState = reading;

if (!on_hold)
{
// Current sampling
 Voltage = AC_VPP();
 VRMS = (Voltage/2.0)*0.707; 
 AmpsRMS = ((VRMS * 1000)/mVperAmp)-0.12; // calibration

// Hack
 if (AmpsRMS < 0.005)
  AmpsRMS=0;

  dtostrf(AmpsRMS, 4, 2, s_amps);                               

 Power = (unsigned int)(Line * AmpsRMS);

// Smoothing

  // subtract the last reading:
  total = total - readings[readIndex];
  // read from the sensor:
  readings[readIndex] = Power;
  // add the reading to the total:
  total = total + readings[readIndex];
  // advance to the next position in the array:
  readIndex = readIndex + 1;

  // if we're at the end of the array...
  if (readIndex >= numReadings) {
    // ...wrap around to the beginning:
    readIndex = 0;
  }

  // calculate the average:
  PowerAVG = total / numReadings;
  last_event = curr_time;
  on_hold = true;
}

// Webserver 
server.handleClient();

#ifdef OTA
// OTA
ArduinoOTA.handle();
#endif
}
